#include <windows.h>
#include <stdio.h>



/*****************************************************************
 *                                                               *
 * The insxmit code, source and everything, is hereby placed in  *
 * the Public Domain. That means you may do with it whatever you *
 * wish, including any alterations, selling, or leasing it. The  *
 * only action you may not take is claim copyright on property   *
 * placed in the Public Domain.                                  *
 *                                                               *
 *                             - Felix Kasza (felixk@dms.shd.de) *
 *                                                               *
 *****************************************************************/



// insxmit.cpp reads a batch file of articles to transmit, strips
// the path header if desired, optionally inserts X- headers
// describing the action, and sends the articles listed in the
// batch through an IHAVE feed to an NNTP server. The format of
// the batch file is: one article per line; each line consists of
// the full path to an article file, followed by whitespace, in
// turn followed by the optional message-id of the article
// (enclosed in angle brackets). If the ID is missing, it is read
// from the article itself. It is assumed that leading dots in the
// article are already doubled; a single "." on a line by itself
// is taken to mean "end of article".



// constants
#define	MAXL	2048 // twice the NNTP legal limit



// these are options and the like
int opt_debugnntp = 0; // debug output if non-0
const char *opt_hostname = "localhost"; // this machine
const char *opt_proto = "nntp"; // service to use
int opt_xheader = 0; // insert X-Insxmit-Date header
int opt_xnewdate = 0; // insert X-Original-Date header, rewrite Date
int opt_newpath = 0; // create a new path
char opt_site[MAXL] = "news!not-for-mail"; // sitename for new path header
int opt_authn = 0; // do authentication if non-0
const char *opt_user = ""; // username for auth'n
const char *opt_pass = ""; // password for auth'n
const char *opt_prefix = ""; // default prefix
char opt_newgrp[MAXL] = ""; // new newsgroups to post to


// article entry
typedef struct entry_
{
	struct entry_ *next; // link ptr
	char *path; // full path
	char *id; // article id
} entry;



// global vars? Who, _me_?
SOCKET sock; // the socket over which we talk
entry *head = NULL, *tail = NULL; // basis of linked list



// function prototypes
int main( int argc, char *argv[] );
void wserr( int rc, const char * const funcname );
int r( char *p ); // read a line from the other guy
void w( const char *const p, int ignoredbg = 0 ); // write a line to the other guy
void readfile( FILE *fp ); // read an article list
void movestuff( void ); // xmit articles
int getl( FILE *fp, char *p ); // read a line from a file
void xmit( SOCKET sock, const entry *const e ); // transmit one article


// getl() reads a line and returns -1 if end of file,
// 0 otherwise
int getl( FILE *fp, char *p )
{
	int c, cnt = 0;

	while ( ( c = getc( fp ) ) != EOF && c != '\n' && c != '\r' )
	{
		if ( c == '\r' )
			continue; // gobble up CRs
		*( p ++ ) = (char) c;
		cnt ++;
		// check line length
		if ( cnt >= MAXL - 2 ) // reserve 1 for '\0', one for doubled dot
			break;
	}
	*p = '\0';

	return c == EOF? EOF: 0;
}



void fixup( char *p )
{
	int l, plen;

	if ( p == NULL || opt_prefix == NULL || *opt_prefix == '\0' )
		return;

	l = strlen( p ) + 1; // include \0 byte
	plen = strlen( opt_prefix ); // don't include term byte

	for ( ; *p; p ++, l -- )
	{
		if ( *p == '<' )
		{
			p ++, l --;
			memmove( p + plen, p, l );
			memcpy( p, opt_prefix, plen );
			l += plen;
		}
	}
}



// xmit() offers an article
void xmit( SOCKET sock, entry *e )
{
	FILE *fp;
	SYSTEMTIME st;
	static const char *const month[] = { "Jan", "Feb", "Mar", "Apr",
		"May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	static const char *const dow[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
	int rc, eof, inheader = 1;
	char l[MAXL * 2], m[MAXL * 2], *p, newid[512], oldid[512];

	fp = fopen( e->path, "rt" );
	if ( fp == NULL )
		return;

	// here, we pre-read (and then rewind) the article to get the
	// message ID. I do this because I am lazy. Ideally, we would
	// pre-read and save the entire header, fix up the headers
	// that need it, and later on write them out, followed by the
	// rest of the article directly from the file.

	while ( 1 )
	{
		eof = getl( fp, l );
		if ( eof && *l != '\0' ) // did we get a line?
			eof = 0; // clear, it will get set next time around
		if ( eof || ( l[0] == '.' && l[1] == '\0' ) ) // end of article?
			break;
		if ( _strnicmp( "Message-ID:", l, 11 ) )
			continue; // no match, try again
		// we found it!
		break;
	}
	p = l + 11;
	while ( *p == ' ' || *p == '\t' )
		++ p;
	strcpy( oldid, p );
	strcpy( newid, p );
	fixup( newid );

	rewind( fp );

	// send IHAVE command, check result
	_snprintf( l, MAXL, "IHAVE %s", newid );
	w( l );
	rc = r( l ); // get reply
	if ( rc < 300 || rc > 399 ) // probably 435, 436, 437
	{
		fclose( fp );
		return;
	}

	while ( 1 )
	{
		eof = getl( fp, l );
		if ( eof && *l != '\0' ) // did we get a line?
			eof = 0; // clear, it will get set next time around
		if ( eof || ( l[0] == '.' && l[1] == '\0' ) ) // end of article?
			break;
		for ( p = &l[0]; *p == ' ' || *p == '\t'; p ++ )
			; // skip whitespace

		if ( inheader && opt_newpath && _strnicmp( "Path:", l, 5 ) == 0 )
		{
			// kill this line, send new path line instead
			w( opt_site, 1 );
			continue;
		}

		if ( inheader && *opt_prefix && _strnicmp( "Message-ID:", l, 11 ) == 0 )
		{
			// kill this line, send new ID instead
			_snprintf( l, sizeof l, "Message-ID: %s", newid );
			w( l, 1 );
			// add old message ID
			_snprintf( l, sizeof l, "X-Insxmit-Message-ID: %s", oldid );
			w( l, 1 );
			continue;
		}

		if ( inheader && *opt_prefix && _strnicmp( "References:", l, 11 ) == 0 )
		{
			// add old message ID
			_snprintf( m, sizeof m, "X-Insxmit-References:%s", l + 11 );
			w( m, 1 );
			// fix up the references and write them out
			fixup( l + 11 );
			w( l, 1 );
			continue;
		}

		if ( inheader && *opt_newgrp && _strnicmp( "Newsgroups:", l, 11 ) == 0 )
		{
			// add old newsgroups
			_snprintf( m, sizeof m, "X-Insxmit-Newsgroups:%s", l + 11 );
			w( m, 1 );
			// fix up the groups and write them out
			w( opt_newgrp, 1 );
			continue;
		}

		if ( inheader && opt_xnewdate && _strnicmp( "Date:", l, 5 ) == 0 )
		{
			memmove( l + 11, l, strlen( l ) + 1 ); // make room
			strncpy( l, "X-Original-", 11 );
			// send modified date line
			w( l, 1 );
			// prepare rewritten date
			GetSystemTime( &st );
			_snprintf( l, MAXL, "Date: %s, %02d %s %04d %02d:%02d:%02d GMT",
				dow[(int) st.wDayOfWeek], (int) st.wDay, month[(int) st.wMonth - 1],
				(int) st.wYear, (int) st.wHour, (int) st.wMinute, (int) st.wSecond );
			w( l, 1 );
			continue;
		}

		if ( inheader && *p == '\0' ) // p points to end of string if line is blank
		{
			inheader = 0;
			if ( opt_xheader )
			{
				// if necessary, send X-Insxmit-Date: header
				GetSystemTime( &st );
				_snprintf( l, MAXL, "X-Insxmit-Date: %s, %02d %s %04d %02d:%02d:%02d GMT",
					dow[(int) st.wDayOfWeek], (int) st.wDay, month[(int) st.wMonth - 1],
					(int) st.wYear, (int) st.wHour, (int) st.wMinute, (int) st.wSecond );
				w( l, 1 );
			}
			// send the blank line
			w( "", 1 );
			continue;
		}

		// send the line as-is
		w( l, 1 );
	}

	fclose( fp );
	w( ".", 1 );
	rc = r( l ); // read result
	// not much we can do about the result
}



// rdc() returns the next character from the socket.
char rdc()
{
	static char buf[MAXL];
	static int nbuf = 0; // chars in buffer
	static char *bufp; // pointer to next byte to return

	while ( nbuf == 0 ) // must fill buffer?
	{
		bufp = buf;
		nbuf = recv( sock, buf, sizeof buf, 0 );
		if ( nbuf == SOCKET_ERROR )
			wserr( nbuf, "recv" );
	}

	nbuf --;
	return *( bufp ++ );
}



// r() reads a line from the socket and returns the value of the
// integer at its start
int r( char *p )
{
	char c;
	char *orig_p = p;

	// gobble up \r, \n
	while ( ( c = rdc() ) == '\r' || c == '\n' )
		;
	// the first byte
	*( p ++ ) = c;
	while ( ( c = rdc() ) != '\r' && c != '\n' )
		*( p ++ ) = c;

	*p = '\0'; // terminate received string

	if ( opt_debugnntp )
	{
		fputs( " -> ", stderr );
		fputs( orig_p, stderr );
		fputs( "\n", stderr );
	}

	return atoi( orig_p );
}



// w() writes a string to the other guy and terminates it with
// carriage-return./linefeed. ignoredbg, if true, suppresses debug output for article lines.
void w( const char *const p, int ignoredbg /* = 0 */ )
{
	static char crlf[] = "\r\n";

	if ( strlen( p ) != (size_t) send( sock, p, strlen( p ), 0 ) )
		wserr( 3, "send" );
	if ( 2 != send( sock, crlf, 2, 0 ) )
		wserr( 4, "send" );

	if ( opt_debugnntp )
	{
		if ( !ignoredbg )
		{
			fputs( "<-  ", stderr );
			fputs( p, stderr );
			fputs( "\n", stderr );
		}
		else
		{
			if ( *p == '.' && p[1] == '\0' ) // article end?
				fputs( "<-  [article]\n", stderr );
		}
	}
}



// wserr() displays winsock errors and aborts. No grace there.
void wserr( int rc, const char * const funcname )
{
	if ( rc == 0 )
		return;

	fprintf( stderr, "\nWinsock error %d [%d] returned by %s().\n"
		"Sorry, no bonus!\n", rc, WSAGetLastError(), funcname );
	WSACleanup();
	exit( rc );
}



// initialize() handles banners, "mode reader", and authentication
void initialize()
{
	char line[MAXL];
	int rc;

	// read the banner line
	rc = r( line );
	if ( rc != 200 )
	{
		fprintf( stderr, "Bad banner: '%s'\n", line );
		exit( 2 );
	}

	if ( opt_authn )
	{
		char cmd[MAXL];

		// send authinfo user
		sprintf( cmd, "authinfo user %s", opt_user );
		w( cmd );
		// wait for 381
		rc = r( line );
		if ( rc != 381 )
		{
			fprintf( stderr, "Authentication (user) failed: '%s'\n", line );
			exit( 2 );
		}

		// send authinfo pass
		sprintf( cmd, "authinfo pass %s", opt_pass );
		w( cmd );
		// wait for 2xx
		rc = r( line );
		if ( rc < 200 || rc > 299 )
		{
			fprintf( stderr, "Authentication (password) failed: '%s'\n", line );
			exit( 2 );
		}
	}
}



// readfile() read a list of paths and article-IDs and adds it
// to the list of articles to send.
void readfile( FILE *fp )
{
	entry *e;
	int n;
	char path[MAXL], id[MAXL], l[MAXL];

	while ( 1 )
	{
loop:
		if ( NULL == fgets( l, sizeof l, fp ) )
			break;

		n = sscanf( l, " %s %s", path, id );

		if ( n < 1 )
			goto loop;

		e = (entry *) malloc( sizeof entry );
		if ( e == NULL )
		{
			fputs( "Panic! out of memory\n", stderr );
			exit( 99 );
		}
		e->next = NULL; // clear link
		e->path = _strdup( path );
		e->id = n >= 2? _strdup( id ): NULL;
		if ( e->path == NULL || ( n >= 2 && e->id == NULL ) )
		{
			fputs( "More panic! out of memory\n", stderr );
			exit( 99 );
		}
		if ( tail == NULL ) // no root node yet?
			head = tail = e;
		else
		{
			tail->next = e;
			tail = e;
		}
	}
}



// in a future version, movestuff() will launch a bundle of worker threads
// (requires that me make the article list and the socket stuff threadsafe)
void movestuff( void )
{
	entry *e;

	while ( head != NULL )
	{
		// remove the item from the queue
		//* MT: lock queue
		// now we have access to the queue, check if someone else grabbed
		// the last entry in the meantime
		if ( head == NULL ) // darn
		{
			//* MT: unlock queue
			break; // exit while loop
		}
		// something was there, go get it
		e = head;
		head = head->next;
		//* MT: unlock queue
		// the item now belongs to us alone
		// xmit article
		xmit( sock, e );
		if ( e->path ) free( e->path );
		if ( e->id ) free( e->id );
		free( e );
	}
}



int main( int argc, char *argv[] )
{
	int rc, i, usage = 0, port;
	char l[MAXL];
	unsigned long naddr;
	FILE *fp;
	WSADATA wsadata;
	PHOSTENT phe;
	PSERVENT pse;
	SOCKADDR_IN addr;

	// parse switches
	if ( argc == 1 )
	{
show_usage:
		puts( "insxmit [options] [file ...]" );
		puts( "" );
		puts( "Options are valid for the entire command. Supported options:" );
		puts( "-d: output all sends and receives on the socket" );
		puts( "-s <server>: the NNTP server's address, either a DNS name or" );
		puts( "    an IP address in dotted-decimal notation. Default: 127.0.0.1" );
		puts( "-p <port>: The port to talk to on the server, as a decimal" );
		puts( "    number or a service as per getservbyname(). Default: 119" );
		puts( "-u <username> <password>: Requires two args, a username and a" );
		puts( "    password, and causes authentication commands to be sent. For" );
		puts( "    now, only clear-text authentication is supported." );
		puts( "-n <site>: strips the existing Path: header and creates a new one." );
		puts( "    The new path will be \"<site>!not-for-mail\"." );
		puts( "-r: rewrites the Date: header, replacing the value with today's" );
		puts( "    date. The original date is saved in X-Original-Date:." );
		puts( "-g <newsgroups>: writes a new Newsgroups: header, inserting the" );
		puts( "    provided string as the newsgroups to post to. If the string" );
		puts( "    contains blanks, it must be enclosed in quotes." );
		puts( "-i <prefix>: creates new, unique, message-IDs by prefixing them" );
		puts( "    with the supplied string. This defeats checks for duplicate" );
		puts( "    articles, so you want to be _very_ careful using -i." );
		puts( "-x: Inserts an X-Insxmit-Date: header into every article." );
		puts( "file: input files, no wildcards, as many as you like. \"-\" causes" );
		puts( "    insxmit to read from stdin." );
		puts( "" );
		puts( "insxmit.cpp reads a batch file of articles to transmit, strips" );
		puts( "the path header if desired, optionally inserts X- headers" );
		puts( "describing the action, and sends the articles listed in the" );
		puts( "batch through an IHAVE feed to an NNTP server. The format of" );
		puts( "the batch file is: one article per line; each line consists of" );
		puts( "the full path to an article file, followed by whitespace, in" );
		puts( "turn followed by the optional message-id of the article" );
		puts( "(enclosed in angle brackets). If the ID is missing, it is read" );
		puts( "from the article itself. It is assumed that leading dots in the" );
		puts( "article are already doubled; a single \".\" on a line by itself" );
		puts( "is taken to mean \"end of article\"." );
		return 1;
	}

	for ( i = 1; i < argc; i ++ )
	{
		if ( *argv[i] != '-' || argv[i][1] == '\0' )
			continue;
		switch ( argv[i][1] )
		{
		case '?':
			goto show_usage;
		case 'd': // debug
			opt_debugnntp ++;
			break;
		case 'n': // new path header
			if ( i >= argc - 1 )
			{
				fputs( "-n requires an argument.\n\n", stderr );
				usage = 1;
				break;
			}
			argv[i] = NULL;
			opt_newpath ++;
			_snprintf( opt_site, sizeof opt_site, "Path: %s!not-for-mail", argv[++ i] );
			break;
		case 'x': // record Insxmit run date
			opt_xheader ++;
			break;
		case 'r': // rewrite Date: header
			opt_xnewdate ++;
			break;
		case 's': // server
			if ( i >= argc - 1 )
			{
				fputs( "-s requires an argument.\n\n", stderr );
				usage = 1;
				break;
			}
			argv[i] = NULL;
			opt_hostname = argv[++ i];
			break;
		case 'p': // port
			if ( i >= argc - 1 )
			{
				fputs( "-p requires an argument.\n\n", stderr );
				usage = 1;
				break;
			}
			argv[i] = NULL;
			opt_proto = argv[++ i];
			break;
		case 'i': // message-ID prefix
			if ( i >= argc - 1 )
			{
				fputs( "-i requires an argument.\n\n", stderr );
				usage = 1;
				break;
			}
			argv[i] = NULL;
			opt_prefix = argv[++ i];
			break;
		case 'g': // new Newsgroups: header
			if ( i >= argc - 1 )
			{
				fputs( "-g requires an argument.\n\n", stderr );
				usage = 1;
				break;
			}
			argv[i] = NULL;
			strcpy( opt_newgrp, "Newsgroups: " );
			strcat( opt_newgrp, argv[++ i] );
			break;
		case 'u': // username, password
			if ( i >= argc - 2 )
			{
				fputs( "-u requires two arguments.\n\n", stderr );
				usage = 1;
				break;
			}
			argv[i] = NULL;
			opt_user = argv[++ i];
			argv[i] = NULL;
			opt_pass = argv[++ i];
			opt_authn = 1;
			break;
		default:
			fprintf( stderr, "Unknown option: '%s'\n", argv[i] );
			usage = 1;
			break;
		}
		argv[i] = NULL;
	}
	if ( usage )
		goto show_usage;

	rc = WSAStartup( 2, &wsadata );
	wserr( rc, "WSAStartup" );

	sock = socket( AF_INET, SOCK_STREAM, 0 );
	if ( sock == INVALID_SOCKET )
		wserr( 999, "socket" );

	addr.sin_family = AF_INET;
	// try numeric IP address first (inet_addr)
	naddr = inet_addr( opt_hostname );
	if ( naddr != INADDR_NONE )
	{
		addr.sin_addr.s_addr = naddr;
	}
	else
	{
		phe = gethostbyname( opt_hostname );
		if ( phe == NULL )
			wserr( 1, "gethostbyname" );
		addr.sin_addr.s_addr = *( (unsigned long *) (phe->h_addr) );
		memcpy( (char *) &addr.sin_addr, phe->h_addr, phe->h_length );
	}

	// try numeric protocol first
	port = atoi( opt_proto );
	if ( port > 0 && port < 23768 )
		addr.sin_port = htons( (short) port );
	else
	{
		pse = getservbyname( opt_proto, "tcp" );
		if ( pse == NULL )
			wserr( 1, "getservbyname" );
		addr.sin_port = pse->s_port;
	}

	rc = connect( sock, (SOCKADDR *) &addr, sizeof addr );
	wserr( rc, "connect" );

	// initialize, authenticate, and so on
	initialize();

	// process input
	for ( i = 1; i < argc; i ++ )
	{
		if ( argv[i] != NULL )
		{
			if ( *argv[i] == '-' && argv[i][1] == '\0' ) // stdin?
				readfile( stdin );
			else
			{
				fp = fopen( argv[i], "rt" );
				if ( fp == NULL )
					fprintf( stderr, "cannot open file, skipping: '%s'\n", argv[i] );
				else
				{
					readfile( fp );
					fclose( fp );
				}
			}
		}
	}

	// file list built, now start transmitting
	movestuff();

	// close connection with QUIT
	w( "quit" );
	rc = r( l );
	if ( rc < 200 || rc > 299 )
		fprintf( stderr, "Bad response to QUIT: '%s'\n", l );

	// close socket
	rc = closesocket( sock );
	wserr( rc, "closesocket" );

	rc = WSACleanup();
	wserr( rc, "WSACleanup" );

	return 0;
}
